/*
 * Copyright (C) 2005-2006 by CERN
 * All rights reserved
 */

#ifndef lint
static char sccsid[] = "@(#)$RCSfile: lcg_globus_ifce.c,v $ $Revision: 1.35 $ $Date: 2009/08/27 09:55:20 $ CERN Jean-Philippe Baud";
#endif /* not lint */

/*      lcg_globus_ifce - wrapper for globus APIs */
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "globus_gass_copy.h"
#include "gfal_api.h"
#include "lcg_util.h"

#ifdef GLOBUS_ERROR_TYPE_MULTIPLE
#define GLOBUS_ERROR2STRING(errorobj) globus_error_print_friendly (errorobj)
#else
#define GLOBUS_ERROR2STRING(errorobj) globus_object_printable_to_string (errorobj)
#endif


typedef struct {
	globus_mutex_t mutex;
	globus_cond_t cond;
	volatile globus_bool_t done;
	volatile globus_bool_t errflag;
	globus_object_t *error;
} monitor_t;

typedef struct {
	monitor_t *mt;
	copyfile_callback_t *user_cb;
	volatile globus_bool_t done;
}  check_t;


static int ggc_interrupted;
static int ggc_interrupt_processed;
static int istimeout;
struct timespec usercb_ts;
perf_t s_transfer_performance;

static volatile int globus_activated = 0;

	static int
do_globus_activates(void)
{
	/* not thread safe */
	if (!globus_activated) {
		globus_activated = 1;
		(void) globus_module_activate (GLOBUS_GASS_COPY_MODULE);
		(void) globus_module_activate (GLOBUS_FTP_CLIENT_MODULE);
	}

	return GLOBUS_SUCCESS;
}

	static int
do_globus_deactivates(void)
{
	if (globus_activated) {
		globus_activated = 0;
		(void) globus_module_deactivate (GLOBUS_GASS_COPY_MODULE);
		(void) globus_module_deactivate (GLOBUS_FTP_CLIENT_MODULE);
	}
	return GLOBUS_SUCCESS;
}



	static int
gftp_client_wait(monitor_t *mp, globus_ftp_client_handle_t *gfhp, const int timeout)
{
	struct timespec ts;
	int save_errno = 0;

	globus_mutex_lock (&mp->mutex);

	if (timeout) {
		ts.tv_sec = time (0) + timeout;
		ts.tv_nsec = 0;
		while (! mp->done && !save_errno) {
			save_errno = globus_cond_timedwait (&mp->cond, &mp->mutex, &ts);
			if (save_errno == EINTR)
				save_errno = 0;
		}
	} else
		while (! mp->done && !save_errno) {
			save_errno = globus_cond_wait (&mp->cond, &mp->mutex);
			if (save_errno == EINTR)
				save_errno = 0;
		}
	if (!mp->done) {
		int ret;
		globus_ftp_client_abort (gfhp);
		do {
			ret = globus_cond_wait (&mp->cond, &mp->mutex);
			if (ret == EINTR)
				ret = 0;
		} while(!mp->done && !ret);
		if (ret && !save_errno)
			save_errno = ret;
	}

	globus_mutex_unlock (&mp->mutex);

	if (save_errno) {
		errno = save_errno;
		return(-1);
	}
	return(0);
}

static int
scan_errstring(const char *p) {
	int ret = ECOMM;

	if (p == NULL) return ret;

	if (strstr (p, "o such file"))
		ret = ENOENT;
	else if (strstr (p, "ermission denied") ||
            strstr (p, "credential"))
		ret = EACCES;
	else if (strstr (p, "exists"))
		ret = EEXIST;
	else if (strstr (p, "command not understood") ||
            strstr (p, "commands denied") ||
            strstr (p, "nsupported checksum"))
		ret = EOPNOTSUPP;

	return ret;
}

	static void
gcallback (void *callback_arg, globus_ftp_client_handle_t *ftp_handle, globus_object_t *error)
{
	monitor_t *monitor = (monitor_t *) callback_arg;
	globus_mutex_lock (&monitor->mutex);
	if (error != GLOBUS_SUCCESS) {
		monitor->errflag = GLOBUS_TRUE;
		monitor->error = globus_object_copy (error);
	}
	monitor->done = GLOBUS_TRUE;
	globus_cond_signal (&monitor->cond);
	globus_mutex_unlock (&monitor->mutex);
}

	static void
ggc_callback (void *callback_arg, globus_gass_copy_handle_t *handle, globus_object_t *error)
{
	monitor_t *monitor = (monitor_t *) callback_arg;
	globus_mutex_lock (&monitor->mutex);
	if (error != GLOBUS_SUCCESS) {
		monitor->errflag = GLOBUS_TRUE;
		monitor->error = globus_object_copy (error);
	}
	monitor->done = GLOBUS_TRUE;
	globus_cond_signal (&monitor->cond);
	globus_mutex_unlock (&monitor->mutex);
}

	static void
ggc_cancel_callback (void *callback_arg, globus_gass_copy_handle_t *handle, globus_object_t *error)
{
	monitor_t *monitor = (monitor_t *) callback_arg;
	globus_mutex_lock (&monitor->mutex);
	if (error != GLOBUS_SUCCESS) {
		monitor->errflag = GLOBUS_TRUE;
		monitor->error = globus_object_copy (error);
	}
	monitor->done = GLOBUS_TRUE;
	globus_cond_signal (&monitor->cond);
	globus_mutex_unlock (&monitor->mutex);
}

static void
gsiftp_copy_performance(
		void *				user_arg,
		globus_gass_copy_handle_t *	handle,
		globus_off_t                    total_bytes,
		float				instantaneous_throughput,
		float				avg_throughput)
{
	s_transfer_performance.transferred_bytes = total_bytes;
	s_transfer_performance.instant_throughput = instantaneous_throughput;
	s_transfer_performance.avg_throughput = avg_throughput;
	usercb_ts.tv_sec = time (0);
	usercb_ts.tv_nsec = 0;
}

void copy_print_callback(perf_t perf, struct timespec ts, int *timeoutflg)
{
	fprintf(stderr,"\r %12lld bytes", (GFAL_LONG64)perf.transferred_bytes);
	fprintf(stderr," %9.2lf KB/sec avg", (double)perf.avg_throughput/1024);
	fprintf(stderr," %9.2lf KB/sec inst", (double)perf.instant_throughput/1024);
}

	static void
timeout_callback(void * user_arg)
{
	check_t * check;
	monitor_t *  monitor;
	copyfile_callback_t *cp_callback;
	globus_reltime_t   delay;
	check = (check_t *)user_arg;
	monitor = (monitor_t *)check->mt;
	cp_callback = (copyfile_callback_t *)check->user_cb;
	globus_mutex_lock(&monitor->mutex);
	{
		GlobusTimeReltimeSet(delay, 1, 0);
		if (*cp_callback)
			(*cp_callback)(s_transfer_performance, usercb_ts, &istimeout);
		if (!monitor->done && !istimeout && !ggc_interrupted){
			globus_callback_register_oneshot(
					NULL,
					&delay,
					timeout_callback,
					check);
		} else
			check->done = GLOBUS_TRUE;
	}
	globus_cond_signal(&monitor->cond);
	globus_mutex_unlock(&monitor->mutex);
}

	int
getfilesize (char *file, globus_off_t *size, char *errbuf, int errbufsz)
{

	return (getfilesizet (file, size, errbuf, errbufsz, 0));

}


	int
getfilesizet (char *file, globus_off_t *size, char *errbuf, int errbufsz, int timeout)
{
	globus_ftp_client_handle_t ftp_handle;
	globus_ftp_client_handleattr_t ftp_handleattr;
	globus_ftp_client_operationattr_t ftp_op_attr;
	globus_result_t gresult;
	monitor_t monitor;
	char *p;
	int save_errno = 0;

	globus_mutex_init (&monitor.mutex, NULL);
	globus_cond_init (&monitor.cond, NULL);
	monitor.done = GLOBUS_FALSE;
	monitor.errflag = GLOBUS_FALSE;
	do_globus_activates();
	globus_ftp_client_handleattr_init (&ftp_handleattr);
	globus_ftp_client_handle_init (&ftp_handle, &ftp_handleattr);
	globus_ftp_client_operationattr_init (&ftp_op_attr);
	gresult = globus_ftp_client_size (&ftp_handle, file, &ftp_op_attr,
			size, &gcallback, &monitor);

	/* check the state immdiately */
	if (gresult != GLOBUS_SUCCESS ) {
		globus_object_t *errobj;
		monitor.errflag = GLOBUS_TRUE;
		errobj = globus_error_get(gresult);
		p = GLOBUS_ERROR2STRING (errobj);
		globus_object_free(errobj);
		goto err;
	}

	if (gftp_client_wait(&monitor, &ftp_handle, timeout))
		save_errno = errno;

	if (monitor.errflag) {
		int err;
		p = GLOBUS_ERROR2STRING (monitor.error);
		globus_object_free (monitor.error);
err:
		err = scan_errstring(p);
		if (err == EEXIST)
			err = EINVAL;

		if (p != NULL) {
			if (err != ENOENT && !save_errno)
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s", p);

			globus_libc_free(p);
		}
		if (!save_errno)
			save_errno = err;
	}

	globus_mutex_destroy(&monitor.mutex);
	globus_cond_destroy(&monitor.cond);
	globus_ftp_client_operationattr_destroy (&ftp_op_attr);
	globus_ftp_client_handle_destroy (&ftp_handle);
	globus_ftp_client_handleattr_destroy (&ftp_handleattr);
	(void) do_globus_deactivates();

	if (monitor.errflag == 0 && !save_errno)
		return (0);

	errno = save_errno;
	return (-1);
}


	int
makedir (char *dir, char *errbuf, int errbufsz)
{
	return (makedirt (dir, errbuf, errbufsz, 0));

}


	int
makedirt (char *dir, char *errbuf, int errbufsz, int timeout)
{
	globus_ftp_client_handle_t ftp_handle;
	globus_ftp_client_handleattr_t ftp_handleattr;
	globus_ftp_client_operationattr_t ftp_op_attr;
	globus_result_t gresult;
	monitor_t monitor;
	char *p;
	int save_errno = 0;

	globus_mutex_init (&monitor.mutex, NULL);
	globus_cond_init (&monitor.cond, NULL);
	monitor.done = GLOBUS_FALSE;
	monitor.errflag = GLOBUS_FALSE;
	do_globus_activates();
	globus_ftp_client_handleattr_init (&ftp_handleattr);
	globus_ftp_client_handle_init (&ftp_handle, &ftp_handleattr);
	globus_ftp_client_operationattr_init (&ftp_op_attr);
	gresult = globus_ftp_client_mkdir (&ftp_handle, dir, &ftp_op_attr,
			&gcallback, &monitor);

	/* check the state immdiately */
	if (gresult != GLOBUS_SUCCESS ) {
		globus_object_t *errobj;
		monitor.errflag = GLOBUS_TRUE;
		errobj = globus_error_get(gresult);
		p = GLOBUS_ERROR2STRING (errobj);
		globus_object_free(errobj);
		goto err;
	}

	if (gftp_client_wait(&monitor, &ftp_handle, timeout))
		save_errno = errno;

	if (monitor.errflag) {
		int err;
		p = GLOBUS_ERROR2STRING (monitor.error);
		globus_object_free (monitor.error);
err:
		err = scan_errstring(p);

		if (p != NULL) {
			if (err != ENOENT && err != EEXIST && !save_errno)
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s", p);

			globus_libc_free(p);
		}
		if (!save_errno)
			save_errno = err;
	}

	globus_mutex_destroy(&monitor.mutex);
	globus_cond_destroy(&monitor.cond);
	globus_ftp_client_operationattr_destroy (&ftp_op_attr);
	globus_ftp_client_handle_destroy (&ftp_handle);
	globus_ftp_client_handleattr_destroy (&ftp_handleattr);
	(void) do_globus_deactivates();

	if (monitor.errflag == 0 && !save_errno)
		return (0);

	errno = save_errno;
	return (-1);
}

	int
makedirp (char *dest_file, char *errbuf, int errbufsz)
{
	return (makedirpt (dest_file, errbuf, errbufsz, 0));

}

	int
makedirpt (char *dest_file, char *errbuf, int errbufsz, int timeout)
{
	int c = 0;
	char *lastslash = NULL;
	char *p;
	char *q;
	char *localbuf;
	int localbufsz;
	int save_errno = 0;

	if (errbuf != NULL) {
		localbufsz = errbufsz;
	} else {
		localbufsz = 8192;
	}
	localbuf = (char *)malloc(localbufsz);
	if (localbuf == NULL)
		return(-1);
	localbuf[0] = '\0';

	p = strrchr (dest_file, '/');
	while (p) {
		if (lastslash == NULL) lastslash = p;
		if (*(p - 1) == ':') break;
		*p = '\0';
		c = makedirt (dest_file, localbuf, localbufsz, timeout);
		if (c && errno == EEXIST) c = 0;
		if (c == 0) {
			*p = '/';
			break;
		}
		if (errno != ENOENT) {
			*p = '/';
			goto ret;
		}
		q = strrchr (dest_file, '/');
		*p = '/';
		p = q;
	}
	while (c == 0 && p < lastslash && (p = strchr (p + 1, '/'))) {
		*p = '\0';
		c = makedirt (dest_file, localbuf, localbufsz, timeout);
		if (c && errno == EEXIST) c = 0;
		*p = '/';
	}
ret:
	save_errno = errno;
	if (c && strlen(localbuf)>0)
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s", localbuf);

	free(localbuf);
	errno = save_errno;

	return (c);
}


	static void
copyfile_sigint_handler(int dummy)
{
	ggc_interrupted = 1;
}

	int
copyfile (char *src_file, char *dest_file, int nbstreams, char *errbuf, int errbufsz)
{

	return (copyfilex (src_file, dest_file, nbstreams, errbuf, errbufsz, 0, 0));
}

	int
copyfilex (char *src_file, char *dest_file, int nbstreams, char *errbuf,
		int errbufsz, int timeout, int verbose)
{
	copyfile_callback_t cp_callback = NULL;
	globus_ftp_control_dcau_t dcau;
	globus_ftp_client_operationattr_t src_ftp_attr,dest_ftp_attr;
	globus_ftp_client_handleattr_t ftp_handleattr;
	globus_gass_copy_attr_t ggc_destattr;
	globus_gass_copy_attr_t ggc_srcattr;
	globus_gass_copy_handle_t ggc_handle;
	globus_gass_copy_handleattr_t ggc_handleattr;
	globus_result_t gresult;
	monitor_t monitor;
	int nretries = 0;
	char *p;
	globus_ftp_control_parallelism_t parallelism;
	struct sigaction sa,oldsa;
	struct timespec ts;
	check_t check;
	int save_errno = 0;

	ggc_interrupted = 0;
	ggc_interrupt_processed = 0;
	istimeout = 0;
	s_transfer_performance.transferred_bytes =0;
	globus_mutex_init (&monitor.mutex, NULL);
	globus_cond_init (&monitor.cond, NULL);
	do_globus_activates();

	// Keep same connection for different gridftp command (cache connections)
	globus_gass_copy_handleattr_init(&ggc_handleattr);
	globus_ftp_client_handleattr_init (&ftp_handleattr);
	globus_ftp_client_handleattr_set_cache_all(&ftp_handleattr, GLOBUS_TRUE);
	globus_gass_copy_handleattr_set_ftp_attr(&ggc_handleattr, &ftp_handleattr);
	globus_gass_copy_handle_init (&ggc_handle, &ggc_handleattr);
	globus_gass_copy_handleattr_destroy (&ggc_handleattr);
	globus_ftp_client_handleattr_destroy (&ftp_handleattr);

	globus_gass_copy_attr_init (&ggc_srcattr);
	globus_gass_copy_attr_init (&ggc_destattr);
	globus_ftp_client_operationattr_init (&src_ftp_attr);
	globus_ftp_client_operationattr_init (&dest_ftp_attr);

	if (strncmp (src_file, "gsiftp:", 7) == 0) {
		if (nbstreams > 1) {
			globus_ftp_client_operationattr_set_mode (&src_ftp_attr,
					GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
			parallelism.mode = GLOBUS_FTP_CONTROL_PARALLELISM_FIXED;
			parallelism.fixed.size = nbstreams;
			globus_ftp_client_operationattr_set_parallelism (&src_ftp_attr,
					&parallelism);
		}
		/* disable data channel authentication */
		dcau.mode = GLOBUS_FTP_CONTROL_DCAU_NONE;
		globus_ftp_client_operationattr_set_dcau (&src_ftp_attr, &dcau);
		globus_gass_copy_attr_set_ftp (&ggc_srcattr, &src_ftp_attr);
	}
	if (strncmp (dest_file, "gsiftp:", 7) == 0) {
		if (nbstreams > 1) {
			globus_ftp_client_operationattr_set_mode (&dest_ftp_attr,
					GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
			parallelism.mode = GLOBUS_FTP_CONTROL_PARALLELISM_FIXED;
			parallelism.fixed.size = nbstreams;
			globus_ftp_client_operationattr_set_parallelism (&dest_ftp_attr,
					&parallelism);
		}
		/* disable data channel authentication */
		dcau.mode = GLOBUS_FTP_CONTROL_DCAU_NONE;
		globus_ftp_client_operationattr_set_dcau (&dest_ftp_attr, &dcau);
		globus_gass_copy_attr_set_ftp (&ggc_destattr, &dest_ftp_attr);
	}
	sa.sa_handler = &copyfile_sigint_handler;
	sigemptyset (&sa.sa_mask);
	sa.sa_flags = 0;
	sigaction (SIGINT, &sa, &oldsa);

	cp_callback = lcg_get_transfer_callback ();
	if (cp_callback == NULL && verbose)
		cp_callback = &copy_print_callback;

retry:
	monitor.done = GLOBUS_FALSE;
	monitor.errflag = GLOBUS_FALSE;
	if (cp_callback != NULL)
		globus_gass_copy_register_performance_cb( &ggc_handle, gsiftp_copy_performance, GLOBUS_NULL);
	gresult = globus_gass_copy_register_url_to_url (&ggc_handle,
			src_file, &ggc_srcattr, dest_file, &ggc_destattr,
			&ggc_callback, &monitor);

	/* check the state immdiately */
	if (gresult != GLOBUS_SUCCESS ) {
		globus_object_t *errobj;
		monitor.errflag = GLOBUS_TRUE;
		errobj = globus_error_get(gresult);
		p = GLOBUS_ERROR2STRING (errobj);
		globus_object_free(errobj);
		goto err;
	}

	globus_mutex_lock (&monitor.mutex);
	check.user_cb = &cp_callback;
	check.mt = &monitor;
	check.done = GLOBUS_FALSE;
	globus_callback_register_oneshot(
			NULL,
			NULL,
			timeout_callback,
			&check);
	if (timeout) {
		ts.tv_sec = time(0) + timeout;
		ts.tv_nsec = 0;
		while ( (! monitor.done) && (!istimeout) && (!save_errno)) {
			if (ggc_interrupted && ! ggc_interrupt_processed) {
				globus_gass_copy_cancel (&ggc_handle,
						&ggc_cancel_callback, &monitor);
				ggc_interrupt_processed = 1;
			}
			save_errno = globus_cond_timedwait (&monitor.cond, &monitor.mutex, &ts);
			if (save_errno == EINTR)
				save_errno = 0;
		}
	} else
		while ( (! monitor.done) && (!istimeout) && (!save_errno)) {
			if (ggc_interrupted && ! ggc_interrupt_processed) {
				globus_gass_copy_cancel (&ggc_handle,
						&ggc_cancel_callback, &monitor);
				ggc_interrupt_processed = 1;
			}
			save_errno = globus_cond_wait (&monitor.cond, &monitor.mutex);
			if (save_errno == EINTR)
				save_errno = 0;
		}
	if (save_errno == ETIMEDOUT)
		istimeout = 1;
	if (!monitor.done){
		int ret;
		globus_gass_copy_cancel (&ggc_handle, &ggc_cancel_callback, &monitor);
		/* wait for the cancel to complete before returning to user */
		do {
			ret = globus_cond_wait (&monitor.cond, &monitor.mutex);
			if (ret == EINTR)
				ret = 0;
		} while(!monitor.done && !ret);
		if (ret && !save_errno)
			save_errno = ret;
	}
	/* wait timeout_callback to complete */
	if (!check.done) {
		int ret;
		do {
			ret = globus_cond_wait (&monitor.cond, &monitor.mutex);
			if (ret == EINTR)
				ret = 0;
		} while(!check.done && !ret);
		if (ret && !save_errno)
			save_errno = ret;
	}
	globus_mutex_unlock (&monitor.mutex);

	if (ggc_interrupt_processed == 1) {
		save_errno = ECANCELED;
	} else if (istimeout == 1) {
		save_errno = ETIMEDOUT;
	}

	if (monitor.errflag) {
		int err;
		p = GLOBUS_ERROR2STRING (monitor.error);
		globus_object_free (monitor.error);
err:
		if (nretries++ == 0 && p != NULL && !save_errno &&
				strstr (p, "o such file") && strncmp (dest_file, "gsiftp:", 7) == 0) {
			/* try to create missing directories:
			   if there is an error we report that rather than the original copy error
			   */
			if (makedirpt (dest_file, errbuf, errbufsz, timeout) == 0) {
				globus_libc_free(p);
				goto retry;
			}
			err = errno;
		} else {
			err = scan_errstring(p);
			if (p != NULL && !save_errno && err != ENOENT && err != EEXIST)
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s", p);
		}

		if (p != NULL)
			globus_libc_free(p);

		if (!save_errno)
			save_errno = err;
	}

	sigaction (SIGINT, &oldsa, NULL);

	globus_mutex_destroy(&monitor.mutex);
	globus_cond_destroy(&monitor.cond);
	globus_gass_copy_handle_destroy (&ggc_handle);
	globus_ftp_client_operationattr_destroy (&src_ftp_attr);
	globus_ftp_client_operationattr_destroy (&dest_ftp_attr);
	do_globus_deactivates();

	if (( monitor.errflag == 0 ) && (!save_errno))
		return 0;

	errno = save_errno;
	return -1;
}


	int
gridftp_checksum (char *file, enum gfal_cksm_type cksmtype, char **cksm,
		int keep_conn, char *errbuf, int errbufsz, int timeout)
{
	char checksum[1024];
	globus_ftp_client_handle_t ftp_handle;
	globus_ftp_client_handleattr_t ftp_handleattr;
	globus_ftp_client_operationattr_t ftp_op_attr;
	globus_result_t gresult;
	monitor_t monitor;
	char *p;
	int save_errno = 0;

	if (file == NULL || cksm == NULL || cksmtype == GFAL_CKSM_NONE) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
				"[GFAL][gridftp_checksum][EINVAL] Invalid arguments");
		errno = EINVAL;
		return (-1);
	}

	memset ((void*)checksum, 0, 1024);

	globus_mutex_init (&monitor.mutex, NULL);
	globus_cond_init (&monitor.cond, NULL);
	monitor.done = GLOBUS_FALSE;
	monitor.errflag = GLOBUS_FALSE;
	do_globus_activates();
	globus_ftp_client_handleattr_init (&ftp_handleattr);

	if (keep_conn)
		/* if keep_conn, tells globus to not close gridftp session */
		globus_ftp_client_handleattr_set_cache_all(&ftp_handleattr, GLOBUS_TRUE);

	globus_ftp_client_handle_init (&ftp_handle, &ftp_handleattr);
	globus_ftp_client_operationattr_init (&ftp_op_attr);
	gresult = globus_ftp_client_cksm (&ftp_handle, file, &ftp_op_attr, checksum,
			0, -1, /* offset, length: we calculate checksum for the whole file */
			gfal_get_checksum_algname (cksmtype), &gcallback, &monitor);

	/* check the state immdiately */
	if (gresult != GLOBUS_SUCCESS ) {
		globus_object_t *errobj;
		monitor.errflag = GLOBUS_TRUE;
		errobj = globus_error_get(gresult);
		p = GLOBUS_ERROR2STRING (errobj);
		globus_object_free(errobj);
		goto err;
	}

	if (gftp_client_wait(&monitor, &ftp_handle, timeout))
		save_errno = errno;

	if (monitor.errflag) {
		int err;
		p = GLOBUS_ERROR2STRING (monitor.error);
		globus_object_free (monitor.error);
err:
		err = scan_errstring(p);
		if (p) globus_libc_free(p);
		if (!save_errno) {
			save_errno = err;
			if (err == EOPNOTSUPP)
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR,
						"%s: CKSM (checksum) operation not supported", file);
		}
	}

	globus_mutex_destroy(&monitor.mutex);
	globus_cond_destroy(&monitor.cond);
	globus_ftp_client_operationattr_destroy (&ftp_op_attr);
	globus_ftp_client_handle_destroy (&ftp_handle);
	globus_ftp_client_handleattr_destroy (&ftp_handleattr);

	if (!keep_conn)
		/* it would close gridftp session */
		do_globus_deactivates();

	if (monitor.errflag == 0 && !save_errno) {
		*cksm = strdup (checksum);
		return (0);
	}

	errno = save_errno;
	return (-1);
}
