#! /bin/bash
#
# Tests for Savannah bug 
#
#   55576: RFE: Extra parameter in lcg-cp for a better TURL construction
#
# https://savannah.cern.ch/bugs/?55576
#

# Global settings. You can override them in the following "edit" section.
source setup.sh

# ++++++++++++ Edit this ++++++++++++
# The remote directory where the test files are placed
TEST_DIR=$GLOBAL_SRM_TEST_DIR_DPM_1
TEST_DIR_FOR_REPLICA=$GLOBAL_SRM_TEST_DIR_DPM_2

# Your VO name
TEST_VO=$GLOBAL_TEST_VO

# The test file that will be copied to the server
LOCAL_TEST_FILE=file:///etc/group
# ++++++++++++ Do not edit after this. ++++++++++++

# Test directly in the build.
PATH=$WS/org.glite.data.dm-util/build/src:$PATH

LCG_CP_COMMAND_BASE="lcg-cp --verbose --nobdii -D srmv2 --vo $TEST_VO"
LCG_DEL_COMMAND_BASE="lcg-del --verbose --nobdii -D srmv2 --vo $TEST_VO"
LCG_LS_COMMAND_BASE="lcg-ls --verbose --nobdii -D srmv2 --vo $TEST_VO"
LCG_CR_COMMAND_BASE="lcg-cr --verbose --nobdii -D srmv2 --vo $TEST_VO -d"
LCG_REP_COMMAND_BASE="lcg-rep --verbose --nobdii -D srmv2 --vo $TEST_VO -d"

REMOTE_FILE_1=$TEST_DIR/testfile_01.dat
REMOTE_FILE_2=$TEST_DIR/testfile_02.dat
REMOTE_FILE_3=$TEST_DIR/testfile_03.dat
REMOTE_FILE_4=$TEST_DIR_FOR_REPLICA/testfile_04.dat

function execute_command {
    local line_number=$1
    local expected_return_code=$2
    local pattern_in_output=$3
    local command=$4
    
    echo -e "\nExecuting command from line $line_number:\n$command\n\n"
    OUTPUT=$(eval $command 2>&1)
    local return_code=$?
    
    if [ $return_code != $expected_return_code ]; then 
        echo -e "Command failed at line $line_number (return code: $return_code, expected: $expected_return_code)"
        echo -e "\nCommand output:\n$OUTPUT\n"
        exit 1
    fi
    
    match=`echo $OUTPUT | grep -c "$pattern_in_output"`
    
    if [ $match == 0 ]; then
        echo -e "Expected pattern \"$pattern_in_output\" cannot be found in the output.\n"
        echo -e "Command output:\n$OUTPUT\n"
        exit 1
    fi 
}

function setup_test {
    execute_command $LINENO 0 "" "$LCG_DEL_COMMAND_BASE -l --force $REMOTE_FILE_1"
    execute_command $LINENO 0 "" "$LCG_DEL_COMMAND_BASE -l --force $REMOTE_FILE_2"
    execute_command $LINENO 0 "" "$LCG_DEL_COMMAND_BASE --force $REMOTE_FILE_3"
    execute_command $LINENO 0 "" "$LCG_DEL_COMMAND_BASE --force $REMOTE_FILE_4"
    execute_command $LINENO 0 "" "$LCG_CP_COMMAND_BASE $LOCAL_TEST_FILE $REMOTE_FILE_1"
    execute_command $LINENO 0 "" "$LCG_CR_COMMAND_BASE $REMOTE_FILE_3 $LOCAL_TEST_FILE"
}

function check_and_remove_file {
    local line_no=$1
    local remote_file=$2
    
    execute_command $line_no 0 "" "$LCG_LS_COMMAND_BASE $remote_file"
    execute_command $line_no 0 "" "$LCG_DEL_COMMAND_BASE -l $remote_file"
}

function check_and_remove_file_with_unreg {
    local line_no=$1
    local remote_file=$2
    
    execute_command $line_no 0 "" "$LCG_LS_COMMAND_BASE $remote_file"
    execute_command $line_no 0 "" "$LCG_DEL_COMMAND_BASE --force $remote_file"
}

function check_if_file_not_exists {
    local line_no=$1
    local remote_file=$2
    
    execute_command $line_no 1 "" "$LCG_LS_COMMAND_BASE $remote_file"
}

function the_tests {
    # REMOTE_FILE_1 has been uploaded. Copy it without specifying protocol list.
    execute_command $LINENO 0 "" "$LCG_CP_COMMAND_BASE $REMOTE_FILE_1 $REMOTE_FILE_2"
    check_and_remove_file $LINENO $REMOTE_FILE_2
    
    # Copy with specifying gsiftp as source protocol only
    execute_command $LINENO 0 "" "$LCG_CP_COMMAND_BASE --src-protocols gsiftp $REMOTE_FILE_1 $REMOTE_FILE_2"
    check_and_remove_file $LINENO $REMOTE_FILE_2
    
    # Copy with specifying gsiftp as destination protocol only
    execute_command $LINENO 0 "" "$LCG_CP_COMMAND_BASE --dst-protocols gsiftp $REMOTE_FILE_1 $REMOTE_FILE_2"
    check_and_remove_file $LINENO $REMOTE_FILE_2
    
    # Copy with specifying gsiftp as source and destination protocol
    execute_command $LINENO 0 "" "$LCG_CP_COMMAND_BASE --src-protocols gsiftp --dst-protocols gsiftp $REMOTE_FILE_1 $REMOTE_FILE_2"
    check_and_remove_file $LINENO $REMOTE_FILE_2
    
    # Copy with specifying gsiftp and an unsupported protocol as source protocol. 
    # Endpoint should select the supported one.
    execute_command $LINENO 0 "" "$LCG_CP_COMMAND_BASE --src-protocols unsupported,gsiftp $REMOTE_FILE_1 $REMOTE_FILE_2"
    check_and_remove_file $LINENO $REMOTE_FILE_2
    
    # Copy with specifying gsiftp and an unsupported protocol as destination protocol. 
    # Endpoint should select the supported one.
    execute_command $LINENO 0 "" "$LCG_CP_COMMAND_BASE --dst-protocols unsupported,gsiftp $REMOTE_FILE_1 $REMOTE_FILE_2"
    check_and_remove_file $LINENO $REMOTE_FILE_2
    
    # Copy with specifying gsiftp and an unsupported protocol as source and destination protocol. 
    # Endpoint should select the supported one.
    execute_command $LINENO 0 "" "$LCG_CP_COMMAND_BASE --src-protocols unsupported,gsiftp --dst-protocols unsupported,gsiftp $REMOTE_FILE_1 $REMOTE_FILE_2"
    check_and_remove_file $LINENO $REMOTE_FILE_2
    
    # Copy with specifying unsupported source protocol. Should return error.
    execute_command $LINENO 1 "Protocol(s) not supported" "$LCG_CP_COMMAND_BASE --src-protocols unsupported $REMOTE_FILE_1 $REMOTE_FILE_2"
    check_if_file_not_exists $LINENO $REMOTE_FILE_2
    
    # Copy with specifying unsupported destination protocol. Should return error.
    execute_command $LINENO 1 "Protocol(s) not supported" "$LCG_CP_COMMAND_BASE --dst-protocols unsupported $REMOTE_FILE_1 $REMOTE_FILE_2"
    check_if_file_not_exists $LINENO $REMOTE_FILE_2
    
    # Copy with specifying unsupported source and destination protocol. Should return error.
    execute_command $LINENO 1 "Protocol(s) not supported" "$LCG_CP_COMMAND_BASE --src-protocols unsupported --dst-protocols unsupported $REMOTE_FILE_1 $REMOTE_FILE_2"
    check_if_file_not_exists $LINENO $REMOTE_FILE_2
    
    # Start testing lcg-cr

    # Copy with specifying gsiftp as destination protocol only
    execute_command $LINENO 0 "" "$LCG_CR_COMMAND_BASE  $REMOTE_FILE_2 --dst-protocols gsiftp $LOCAL_TEST_FILE"
    check_and_remove_file_with_unreg $LINENO $REMOTE_FILE_2
    
    # Copy with specifying gsiftp and an unsupported protocol as source and destination protocol. 
    execute_command $LINENO 0 "" "$LCG_CR_COMMAND_BASE  $REMOTE_FILE_2 --dst-protocols unsupported,gsiftp $LOCAL_TEST_FILE"
    check_and_remove_file_with_unreg $LINENO $REMOTE_FILE_2
    
    # Copy with specifying unsupported destination protocol. Should return error.
    execute_command $LINENO 1 "Protocol(s) not supported" "$LCG_CR_COMMAND_BASE  $REMOTE_FILE_2 --dst-protocols unsupported $LOCAL_TEST_FILE"    
    check_if_file_not_exists $LINENO $REMOTE_FILE_2

    # Start testing lcg-rep
    
    # Copy with specifying gsiftp as source protocol only
    execute_command $LINENO 0 "" "$LCG_REP_COMMAND_BASE $REMOTE_FILE_4 --src-protocols gsiftp $REMOTE_FILE_3"
    check_and_remove_file_with_unreg $LINENO $REMOTE_FILE_4

    # Copy with specifying gsiftp as destination protocol only
    execute_command $LINENO 0 "" "$LCG_REP_COMMAND_BASE $REMOTE_FILE_4 --dst-protocols gsiftp $REMOTE_FILE_3"
    check_and_remove_file_with_unreg $LINENO $REMOTE_FILE_4
    
    # Copy with specifying gsiftp as source and destination protocol
    execute_command $LINENO 0 "" "$LCG_REP_COMMAND_BASE $REMOTE_FILE_4 --src-protocols gsiftp --dst-protocols gsiftp $REMOTE_FILE_3"
    check_and_remove_file_with_unreg $LINENO $REMOTE_FILE_4
    
    # Copy with specifying gsiftp and an unsupported protocol as source protocol. 
    # Endpoint should select the supported one.
    execute_command $LINENO 0 "" "$LCG_REP_COMMAND_BASE $REMOTE_FILE_4 --src-protocols unsupported,gsiftp $REMOTE_FILE_3"
    check_and_remove_file_with_unreg $LINENO $REMOTE_FILE_4
    
    # Copy with specifying gsiftp and an unsupported protocol as destination protocol. 
    # Endpoint should select the supported one.
    execute_command $LINENO 0 "" "$LCG_REP_COMMAND_BASE $REMOTE_FILE_4 --dst-protocols unsupported,gsiftp $REMOTE_FILE_3"
    check_and_remove_file_with_unreg $LINENO $REMOTE_FILE_4
    
    # Copy with specifying gsiftp and an unsupported protocol as source and destination protocol. 
    # Endpoint should select the supported one.
    execute_command $LINENO 0 "" "$LCG_REP_COMMAND_BASE $REMOTE_FILE_4 --src-protocols unsupported,gsiftp --dst-protocols unsupported,gsiftp $REMOTE_FILE_3"
    check_and_remove_file_with_unreg $LINENO $REMOTE_FILE_4
    
    # Copy with specifying unsupported source protocol. Should return error.
    execute_command $LINENO 1 "Protocol(s) not supported" "$LCG_REP_COMMAND_BASE $REMOTE_FILE_4 --src-protocols unsupported $REMOTE_FILE_3"
    check_if_file_not_exists $LINENO $REMOTE_FILE_4
    
    # Copy with specifying unsupported destination protocol. Should return error.
    execute_command $LINENO 1 "Protocol(s) not supported" "$LCG_REP_COMMAND_BASE $REMOTE_FILE_4 --dst-protocols unsupported $REMOTE_FILE_3"
    check_if_file_not_exists $LINENO $REMOTE_FILE_4
    
    # Copy with specifying unsupported source and destination protocol. Should return error.
    execute_command $LINENO 1 "Protocol(s) not supported" "$LCG_REP_COMMAND_BASE $REMOTE_FILE_4 --src-protocols unsupported --dst-protocols unsupported $REMOTE_FILE_3"
    check_if_file_not_exists $LINENO $REMOTE_FILE_4
}

echo -e "\nStarting test $0\n"
setup_test
the_tests
echo -e "\nTest $0 passed\n"
exit 0


